/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                        Human.c                          **/
/**              based on MSDOS.c and Unix.c                **/
/**                                                         **/
/** This file contains Human68k-dependent subroutines and   **/
/** drivers. It includes common drivers from Common.h.      **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin  1994,1995                **/
/**               Igor Sharfmesser 1994,1995                **/
/**               Alex Krasivsky   1995                     **/
/**               Yasushi Yamasaki 1995                     **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, Marat Fayzullin,   **/
/**     if you make any changes to this file.               **/
/*************************************************************/

/** Standard Human68k #includes *********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>

#include <iocslib.h>
#include <doslib.h>

#include "MSX.h"
#include "X68k.h"

/** Various variables ****************************************/
#define WIDTH  320
#define HEIGHT 240

char *Title="fMSX Human68k 0.9";

#define TIMER_COUNT 1

int ExitFlag=0;
int IFreq=50;

byte *XBuf,*SBuf,XPal[16],XPal0,Pal[16][3];

volatile int CntInt;

volatile byte XKeyMap[16] =
{
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF
};

void (*KeyIntrOld)();
void (*TimerIntrOld)();

/** PUTIMAGE: macro copying image buffer into a window *******/
void PutImage256( void );
void PutImage512( void );

#define PutImage PutImage256()


void KeyIntrHandler( void );
void TimerIntrHandler( void );

int InitScr256(void);
int InitScr512(void);


/** Screen Mode Handlers [number of screens + 1] *************/
struct
{                                  /* Screen mode handlers:  */
  int (*Init)();                   /* Initialize screen      */
  void (*Colors)();                /* Update colors/palette  */
  void (*Refresh)(byte Y1,byte Y2);/* Refresh screen         */
  byte R2,R3,R4,R5;                /* Masks for table addr.  */
} SCR[MAXSCREEN+2] =               /* Number of screens + 1  */
{
  { InitScr256,ColorsScrF,RefreshScr0,0x7F,0x00,0x3F,0x00 }, /* SCR 0:TEXT 40x24  */
  { InitScr256,ColorsScrF,RefreshScr1,0x7F,0xFF,0x3F,0xFF }, /* SCR 1:TEXT 32x24  */
  { InitScr256,ColorsScrF,RefreshScr2,0x7F,0x80,0x3C,0xFF }, /* SCR 2:BLK 256x192 */
  { InitScr256,ColorsScrF,RefreshScr3,0x7F,0xFF,0x00,0xFF }, /* SCR 3:64x48x16    */
  { InitScr256,ColorsScrF,RefreshScr4,0x7F,0x80,0x3C,0xFF }, /* SCR 4:BLK 256x192 */
  { InitScr256,ColorsScrF,RefreshScr5,0x60,0x00,0x00,0xFC }, /* SCR 5:256x192x16  */
  { InitScr512,ColorsScrF,RefreshScr6,0x60,0x00,0x00,0xFC }, /* SCR 6:512x192x4   */
  { InitScr512,ColorsScrF,RefreshScr7,0x20,0x00,0x00,0xFC }, /* SCR 7:512x192x16  */
  { InitScr256,ColorsScrF,RefreshScr8,0x20,0x00,0x00,0xFC }, /* SCR 8:256x192x256 */
  { InitScr512,ColorsScrF,RefreshTx80,0x7C,0xF8,0x3F,0x00 }  /* SCR 0:TEXT 80x24  */
};


/** Keyboard bindings ****************************************/
byte Keys[][2] =
{
/*00            ESC       1         2         3         4         5         6     */
  {-1,0x00},{ 7,0x04},{ 0,0x02},{ 0,0x04},{ 0,0x08},{ 0,0x10},{ 0,0x20},{ 0,0x40},
/*08  7         8         9         0         -         ^         \         BS   */
  { 0,0x80},{ 1,0x01},{ 1,0x02},{ 0,0x01},{ 1,0x04},{ 1,0x08},{ 1,0x10},{ 7,0x20},
/*10  Tab       Q         W         E         R         T         Y         U     */
  { 7,0x08},{ 4,0x40},{ 5,0x10},{ 3,0x04},{ 4,0x80},{ 5,0x02},{ 5,0x40},{ 5,0x04},
/*18  I         O         P         @         [         Enter     A         S     */
  { 3,0x40},{ 4,0x10},{ 4,0x20},{ 1,0x20},{ 1,0x40},{ 7,0x80},{ 2,0x40},{ 5,0x01},
/*20  D         F         G         H         J         K         L         ;     */
  { 3,0x02},{ 3,0x08},{ 3,0x10},{ 3,0x20},{ 3,0x80},{ 4,0x01},{ 4,0x02},{ 1,0x80},
/*28  :         ]         Z         X         C         V         B         N     */
  { 2,0x01},{ 2,0x02},{ 5,0x80},{ 5,0x20},{ 3,0x01},{ 5,0x08},{ 2,0x80},{ 4,0x08},
/*30  M         ,         .         /         _         Space     Home      Del   */
  { 4,0x04},{ 2,0x04},{ 2,0x08},{ 2,0x10},{ 2,0x20},{ 8,0x01},{ 8,0x02},{ 8,0x08},
/*38* Rollup    Rolldown  Undo      Left      Up        Right     Down      CLR   */
  {-1,0x00},{-1,0x00},{-1,0x00},{ 8,0x10},{ 8,0x20},{ 8,0x80},{ 8,0x40},{-1,0x00},
/*40  [/]       [*]       [-]       [7]       [8]       [9]       [+]       [4]   */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*48  [5]       [6]       [=]       [1]       [2]       [3]       [Enter]   [0]   */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*50  [,]       [.]       Kigo      Toroku    Help      XF1       XF2       XF3   */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{ 6,0x04},{-1,0x00},{-1,0x00},
/*58  XF4       XF5       Kana      Rome      Code      Caps      Ins       Hira  */
  { 7,0x40},{ 8,0x04},{ 6,0x10},{-1,0x00},{-1,0x00},{ 6,0x08},{ 8,0x04},{-1,0x00},
/*60  Zenkaku   Break     Copy      F1        F2        F3        F4        F5    */
  {-1,0x00},{ 7,0x10},{-1,0x00},{ 6,0x20},{ 6,0x40},{ 6,0x80},{ 7,0x01},{ 7,0x02},
/*68  F6        F7        F8        F9        F10                                 */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*70  Shift     Ctrl    */
  { 6,0x01},{ 6,0x02},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*78*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00}
};


/** A handler for the timer interrupt ************************/
void TimerIntrSub(void)
{
  if(!CntInt--) { CntInt=TIMER_COUNT;IFlag=1; }
}


/** A keyboard interrupt handler subroutine *********************/
void KeyIntrSub(void)
{
  register unsigned char key;

  switch( key = *usart )
  {
    case 0xd4: ExitFlag=1;break;     /* HELP   released */
    case 0xd3: break;                /* TOROKU released */
#ifdef DEBUG
    case 0xd2: Trace=!Trace;break;   /* KIGO   released */
#endif
    default:
      if( Keys[key&0x7f][0] != (byte)-1 )
        if( key&0x80 )
          XKeyMap[Keys[key&0x7f][0]] |= Keys[key&0x7f][1];
        else
          XKeyMap[Keys[key][0]] &= ~Keys[key][1];
  }
}


/** Keyboard *************************************************/
/** Check for keys and modify MSX keyboard matrix.          **/
/*************************************************************/
void Keyboard(void)
{
  memcpy(KeyMap,(byte *)XKeyMap,16); 
  if(ExitFlag)
    { TrashMachine();TrashMSX();exit(0); }
}


/** InitMachine **********************************************/
/** Allocate resources needed by MSDOS-dependent code.      **/
/*************************************************************/
int InitMachine(void)
{
  int I,J;
  static byte PalInit[16][3] =
  {
    {0x00,0x00,0x00},{0x00,0x00,0x00},{0x00,0xA0,0x00},{0x60,0xE0,0x60},
    {0x00,0x00,0x90},{0x00,0x00,0xF0},{0xA0,0x00,0x00},{0x00,0xC0,0xE0},
    {0xE0,0x40,0x40},{0xF0,0x80,0x80},{0xC0,0xA0,0x00},{0xE0,0xE0,0x60},
    {0x00,0x60,0x00},{0xE0,0x60,0xE0},{0x80,0x80,0x80},{0xE0,0xE0,0xE0},
  };

  if(Verbose)
    printf("Initializing Human68k drivers:\n  Allocating screen buffer...");   

  XBuf=(byte *)MALLOC(sizeof(byte)*640*240);
  if( XBuf < 0 ) { if(Verbose) printf("FAILED\n");return(0); }
  if(Verbose) printf("OK\n  Allocating sprite buffer...");
  SBuf=(byte *)MALLOC(sizeof(byte)*256*212);
  if( SBuf < 0 ) { if(Verbose) printf("FAILED\n");return(0); }
  if(Verbose) printf("OK\n");

  /* set hardware timer for interrupt */
  if(!IntSync)
  {
    if((IFreq<2)||(IFreq>60)) IFreq=50;
    if(Verbose) printf("  Initializing timer for about %dHz...",60 / (60 / IFreq));

    CntInt = TIMER_COUNT;
    if( VDISPST( TimerIntrHandler, 0, 60 / IFreq ) != 0 ) {
      printf( "VDISP already used.\n" );exit(1);
    }
  }

  /** set keyboard interrupt handler **/
  KeyIntrOld = INTVCS( 0x4c, KeyIntrHandler );

  C_CUROFF();
  B_SUPER(0);

  /*** Set palette ***/
  for(I=0;I<16;I++) {
    XPal[I]=I;
    gpalet[I] = PalInit[I][1]>>3<<11 | PalInit[I][0]>>3<<6 | PalInit[I][2]>>3<<1;
  }
  XPal0=XPal[0];

  return(1);
}


/** TrashMachine *********************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
  C_WIDTH( 0 );
  C_FNKMOD( 0 );
  C_CURON();

  if(Verbose) printf("Shutting down...\n");

  if(SBuf) free(SBuf);
  if(XBuf) free(XBuf);

  /* recover interrupt vectors */
  INTVCS( 0x4c, KeyIntrOld );
  VDISPST( (void *)0, 0, 0 );
}


/*
 * change crt mode with specified param.
 */
int ChangeCrtMode( struct Crtc *crtcVal,	/* crtc register value */
		   unsigned short vr0,		/* video con. r0 - r2 */
		   unsigned short vr1,
		   unsigned short vr2,
		   unsigned short contrast,	/* crt contrast level */
		   unsigned short dotClk )	/* dot clock selector */
{
	unsigned short r20;

	sysport->r1 = contrast;	/* set screen contrast */
	sysport->r4 = dotClk;	/* set dot clock */

	*videoR0 = vr0;
	*videoR1 = vr1;
	*videoR2 = vr2;

	/* set crt mode resisters */
#define CRTCREG1TO7 \
		crtc->r01 = crtcVal->r01;\
		crtc->r02 = crtcVal->r02;\
		crtc->r03 = crtcVal->r03;\
		crtc->r04 = crtcVal->r04;\
		crtc->r05 = crtcVal->r05;\
		crtc->r06 = crtcVal->r06;\
		crtc->r07 = crtcVal->r07;

	r20 = crtc->r20;
	if( ((r20 & 0x03)<<5 | (r20 & 0x10)) >
	    ((crtcVal->r20 & 0x03)<<5 | (crtcVal->r20 & 0x10)) ) {
		/* change to low resolution mode */
		crtc->r20 = crtcVal->r20;
		CRTCREG1TO7
		crtc->r00 = crtcVal->r00;
	} else {
		/* change to high resolution mode */
		crtc->r00 = crtcVal->r00;
		CRTCREG1TO7
		crtc->r20 = crtcVal->r20;
	}
	/* set horizontal position adjustment param. */
	crtc->r08 = crtcVal->r08;

	/* set scroll registers */
	crtc->r12 = crtcVal->r12;
	crtc->r13 = crtcVal->r13;
	crtc->r14 = crtcVal->r14;
	crtc->r15 = crtcVal->r15;
	crtc->r16 = crtcVal->r16;
	crtc->r17 = crtcVal->r17;
	crtc->r18 = crtcVal->r18;
	crtc->r19 = crtcVal->r19;
}


/** InitScr256 ***********************************************/
/** Function to be called when 256-wide screen is set.      **/
/*************************************************************/
int InitScr256(void)
{
	/* 384 wide screen data */
	static struct Crtc crtc384 = {
		0x0044,0x0003,0x0009,0x0039,0x0237,0x0005,0x0028,0x0228,
		0x001b,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
		0x0001,0x0000,0x0001,0x0000,0x0311
	};

	/* change crt mode to 384x256 */
/*	ChangeCrtMode( &crtc384, 0x0001, 0x06e4, 0x002f, 0x000e, 0xfffe ); */
/* do not display text plane */
	ChangeCrtMode( &crtc384, 0x0001, 0x06e4, 0x000f, 0x000e, 0xfffe );
	memset((unsigned short *)gframe,0,512*256*2);
	return 1;
}

/** ColorScrF ************************************************/
/** Generic function to be called when colors are changed.  **/
/*************************************************************/
void ColorsScrF(void)
{
  register byte C;
  register int I;

  for(I=0;I<16;I++) {
	gpalet[I] = Palette[I][1]>>3<<11 | Palette[I][0]>>3<<6 | Palette[I][2]>>3<<1 ;
  }

  C=XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];
  memset(XBuf,C,640*240);
}

/** InitScr512 ***********************************************/
/** Function to be called when 512-wide screen is set.      **/
/*************************************************************/
int InitScr512(void)
{
	/* 768 wide screen data */
	static struct Crtc crtc768 = {
		0x0089,0x000e,0x001c,0x007c,0x0237,0x0005,0x0028,0x0228,
		0x001b,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,
		0x0000,0x0000,0x0000,0x0000,0x0412
	};

	/* change crt mode to 768x256 */
/*	ChangeCrtMode( &crtc768, 0x0004, 0x06e4, 0x0030, 0x000e, 0xfffc ); */
/* do not display text plane */
	ChangeCrtMode( &crtc768, 0x0004, 0x06e4, 0x0010, 0x000e, 0xfffc );
	memset((unsigned short *)gframe,0,1024*256*2);
	return 1;
}


/** Part of the code common for Unix/X and MSDOS and Human68k drivers *****/
#include "Common.h"


/** RefreshScr6 **********************************************/
/** Function to be called to update SCREEN 6.               **/
/*************************************************************/
void RefreshScr6(register byte Y1,register byte Y2)
{
  register byte X,J,C;
  register byte *P,*T,*R;

  P=XBuf+640*(HEIGHT-192)/2+(640-512)/2+Y1*640;

  if(ScreenON)
  {
    ColorSprites();

    R=SBuf+((word)Y1<<8);
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<7);

    for(;Y1<=Y2;Y1++,J++,P+=(640-512))
    {
      if(!J) T=ChrTab;
      for(X=0;X<128;X++,T++)
      {
        C=((*R>>4)|*R)&0x0F;
        *P++=XPal[C? C:(*T>>6)&0x03];
        *P++=XPal[C? C:(*T>>4)&0x03];
        R++;
        C=((*R>>4)|*R)&0x0F;
        *P++=XPal[C? C:(*T>>2)&0x03];
        *P++=XPal[C? C:*T&0x03];
        R++;
      }
    }
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH);

  if(EndOfFrame) PutImage512();
}


/** RefreshScr7 **********************************************/
/** Function to be called to update SCREEN 7.               **/
/*************************************************************/
void RefreshScr7(register byte Y1,register byte Y2)
{
  register byte X,J,C;
  register byte *P,*T,*R;

  P=XBuf+640*(HEIGHT-192)/2+(640-512)/2+Y1*640;

  if(ScreenON)
  {
    ColorSprites();

    R=SBuf+((word)Y1<<8);
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<8);

    for(;Y1<=Y2;Y1++,J++,P+=(640-512))
    {
      if(!J) T=ChrTab;
      for(X=0;X<128;X++)
      {
        C=((*R>>4)|*R)&0x0F;
        *P++=XPal[C? C:*T>>4];
        *P++=XPal[C? C:*T&0x0F];
        T++;R++;
        C=((*R>>4)|*R)&0x0F;
        *P++=XPal[C? C:*T>>4];;
        *P++=XPal[C? C:*T&0x0F];;
        T++;R++;
      }
    }
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH);

  if(EndOfFrame) PutImage512();
}


/** RefreshScr8 **********************************************/
/** Function to be called to update SCREEN 8.               **/
/*************************************************************/
void RefreshScr8(register byte Y1,register byte Y2)
{
  register byte *P,*T,J;
  register word X;

  P=XBuf+WIDTH*(HEIGHT-192)/2+Y1*WIDTH;

  if(ScreenON)
  {
    P+=(WIDTH-256)/2;
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<8);

    for(;Y1<=Y2;Y1++,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<256;X++,T++) *P++=*T;
    }
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH);

  if(EndOfFrame) PutImage;
}


/** RefreshTx80 **********************************************/
/** Function to be called to update TEXT80.                 **/
/*************************************************************/
void RefreshTx80(register byte Y1,register byte Y2) 
{
  register byte X,Y,J,K,M,FC,BC;
  register byte *P,*S,*C,*T;
  byte *G;

  P=XBuf+640*(HEIGHT-192)/2 + (640-512)/2;
  G=FontBuf? FontBuf:ChrGen;

  if(ScreenON)
  {
    C=ColTab;T=ChrTab;M=0x80;

    for(Y=ScanLines212? 26:24;Y;Y--,P+=640*8-6*80)
      for(X=0;X<80;X++,T++)
      {
        S=G+((long)*T<<3);
        if(*C&M) { FC=XPal[XFGColor];BC=XPal[XBGColor]; }
        else     { FC=XPal[FGColor];BC=XPal[BGColor]; }
        for(J=0;J<8;J++)
        {
          K=*S++;
          *P++=K&0x80? FC:BC; *P++=K&0x40? FC:BC;
          *P++=K&0x20? FC:BC; *P++=K&0x10? FC:BC;
          *P++=K&0x08? FC:BC; *P  =K&0x04? FC:BC;
          P+=640-5;
        }
        P+=6-640*8;
        if(!(M>>=1)) { M=0x80;C++; } 
      }
  }
  else memset(P+((word)Y1<<8),XPal[BGColor],(Y2-Y1+1)*640);
    
  if(EndOfFrame) PutImage512();
}

/*
 * malloc
 */
void *malloc( size_t size )
{
	void *ret;
	if( (ret = MALLOC( size )) < 0 )
		return NULL;
	return ret;
}

/*
 * free
 */
void free( void *blk )
{
	MFREE( blk );
}


