/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                          MSDOS.c                        **/
/**                                                         **/
/** This file contains MSDOS-dependent subroutines and      **/
/** drivers. It includes common drivers from Common.h.      **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin  1994,1995                **/
/**               Igor Sharfmesser 1994,1995                **/
/**               Alex Krasivsky   1995                     **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, Marat Fayzullin,   **/
/**     if you make any changes to this file.               **/
/*************************************************************/


/** Standard MSDOS #includes *********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <conio.h>
#include <i86.h>

#include "MSX.h"


/** Various variables ****************************************/
#define WIDTH  320
#define HEIGHT 240

#pragma aux cli="cli";
#pragma aux sti="sti";

char *Title="fMSX MSDOS 0.9";

int ExitFlag=0;
int IFreq=50;

byte *XBuf,*SBuf,XPal[16],XPal0,Pal[16][3];

volatile int Cmp,CntInt;
volatile long RealClock;

volatile byte XKeyMap[16] =
{
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
  0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF
};

extern void cli();
extern void sti();

void (__interrupt __far *IRQ0Old)();
void (__interrupt __far *IRQ1Old)();


/** PUTIMAGE: macro copying image buffer into a window *******/
#define PutImage \
  memcpy((char *)0xA0000,(char *)XBuf+WIDTH*(HEIGHT-200)/2,320*200)


void Put640x480(byte *Buf);
void InitScr256(void);
void InitScr512(void);


/** Screen Mode Handlers [number of screens + 1] *************/
struct
{                                  /* Screen mode handlers:  */
  int (*Init)();                   /* Initialize screen      */
  void (*Colors)();                /* Update colors/palette  */
  void (*Refresh)(byte Y1,byte Y2);/* Refresh screen         */
  byte R2,R3,R4,R5;                /* Masks for table addr.  */
} SCR[MAXSCREEN+2] =               /* Number of screens + 1  */
{
  { InitScr256,ColorsScrF,RefreshScr0,0x7F,0x00,0x3F,0x00 }, /* SCR 0:TEXT 40x24  */
  { InitScr256,ColorsScrF,RefreshScr1,0x7F,0xFF,0x3F,0xFF }, /* SCR 1:TEXT 32x24  */
  { InitScr256,ColorsScrF,RefreshScr2,0x7F,0x80,0x3C,0xFF }, /* SCR 2:BLK 256x192 */
  { InitScr256,ColorsScrF,RefreshScr3,0x7F,0xFF,0x00,0xFF }, /* SCR 3:64x48x16    */
  { InitScr256,ColorsScrF,RefreshScr4,0x7F,0x80,0x3C,0xFF }, /* SCR 4:BLK 256x192 */
  { InitScr256,ColorsScrF,RefreshScr5,0x60,0x00,0x00,0xFC }, /* SCR 5:256x192x16  */
  { InitScr512,ColorsScrF,RefreshScr6,0x60,0x00,0x00,0xFC }, /* SCR 6:512x192x4   */
  { InitScr512,ColorsScrF,RefreshScr7,0x20,0x00,0x00,0xFC }, /* SCR 7:512x192x16  */
  { InitScr256,ColorsScrF,RefreshScr8,0x20,0x00,0x00,0xFC }, /* SCR 8:256x192x256 */
  { InitTx80,  ColorsTx80,RefreshTx80,0x7C,0xF8,0x3F,0x00 }  /* SCR 0:TEXT 80x24  */
};


/** Keyboard bindings ****************************************/
byte Keys[][2] =
{
/*00            ESC       1         2         3         4         5         6     */
  {-1,0x00},{ 7,0x04},{ 0,0x02},{ 0,0x04},{ 0,0x08},{ 0,0x10},{ 0,0x20},{ 0,0x40},
/*08  7         8         9         0         -         +         BS        Tab   */
  { 0,0x80},{ 1,0x01},{ 1,0x02},{ 0,0x01},{ 1,0x04},{ 1,0x08},{ 7,0x20},{ 7,0x08},
/*10  Q         W         E         R         T         Y         U         I     */
  { 4,0x40},{ 5,0x10},{ 3,0x04},{ 4,0x80},{ 5,0x02},{ 5,0x40},{ 5,0x04},{ 3,0x40},
/*18  O         P         [         ]         ENTER     Ctrl      A         S     */
  { 4,0x10},{ 4,0x20},{ 1,0x20},{ 1,0x40},{ 7,0x80},{ 6,0x02},{ 2,0x40},{ 5,0x01},
/*20  D         F         G         H         J         K         L         ;     */
  { 3,0x02},{ 3,0x08},{ 3,0x10},{ 3,0x20},{ 3,0x80},{ 4,0x01},{ 4,0x02},{ 1,0x80},
/*28  "         `         ShiftL    \         Z         X         C         V     */
  { 2,0x01},{ 2,0x02},{ 6,0x01},{ 1,0x10},{ 5,0x80},{ 5,0x20},{ 3,0x01},{ 5,0x08},
/*30  B         N         M         ,         .         /         ShiftR    [*]   */
  { 2,0x80},{ 4,0x08},{ 4,0x04},{ 2,0x04},{ 2,0x08},{ 2,0x10},{ 6,0x01},{-1,0x00},
/*38* AltL      Space     CapsLock  F1        F2        F3        F4        F5    */
  { 6,0x10},{ 8,0x01},{ 6,0x08},{ 6,0x20},{ 6,0x40},{ 6,0x80},{ 7,0x01},{ 7,0x02},
/*40  F6        F7        F8        F9        F10       NumLock   ScrLock   [7]   */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{ 2,0x20},{-1,0x00},{-1,0x00},
/*48  [8]       [9]       [-]       [4]       [5]       [6]       [+]       [1]   */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*50  [2]       [3]       [0]       [.]     Alt+PrtScr                      F11   */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*58  F12                                                                         */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*60*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*68*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*70*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*78*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*80 Keymap with E0 prefixes */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*08*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*10*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*18                                        [Enter]     CtrlR                     */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{ 6,0x02},{-1,0x00},{-1,0x00},
/*20*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*28                     Shift?                                                   */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*30                                                      [/]              PrtScr */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*38   AltR                                                                       */
  { 6,0x04},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*40                                                   Pause? Ctrl+Break     Home */
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{ 8,0x02},
/*48     Up      PgUp                Left               Right                 End */
  { 8,0x20},{ 7,0x10},{-1,0x00},{ 8,0x10},{-1,0x00},{ 8,0x80},{-1,0x00},{ 7,0x40},
/*50   Down      PgDn       Ins       Del                                         */
  { 8,0x40},{-1,0x00},{ 8,0x04},{ 8,0x08},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*58*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*60*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*68*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*70*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},
/*78*/
  {-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00},{-1,0x00}
};


/** A handler for the timer interrupt ************************/
void __interrupt IRQ0Handler(void)
{
  cli();

  if(!CntInt--) { CntInt=100;IFlag=1; }
  RealClock+=Cmp;
  if(RealClock>0xFFFF) { RealClock&=0xFFFF;IRQ0Old(); }

  outp(0x20,0x20);
  sti();
}


/** A handler for the keyboard interrupt *********************/
void __interrupt IRQ1Handler(void)
{
  static int Offset=0;
  register int Key;
  
  cli();
  
  switch(Key=inp(0x60))
  {
    case 0xD8: ExitFlag=1;break;     /* F12 released */
    case 0xD7: break;                /* F11 released */
#ifdef DEBUG
    case 0xC4: Trace=!Trace;break;   /* F10 released */
#endif
    case 0xE0: Offset+=128;break;
    default:   
      if(Keys[Key&0x7f][0]!=-1)
        if(Key&0x80)
          XKeyMap[Keys[Key&0x7f+Offset][0]]|=Keys[Key&0x7f+Offset][1];
        else
          XKeyMap[Keys[Key+Offset][0]]&=~Keys[Key+Offset][1];
      Offset=0;
  }

  outp(0x20,0x20);
  sti();
}


/** Keyboard *************************************************/
/** Check for keys and modify MSX keyboard matrix.          **/
/*************************************************************/
void Keyboard(void)
{
  memcpy(KeyMap,XKeyMap,16); 
  if(ExitFlag)
  { TrashMachine();TrashMSX();exit(0); }
}


/** InitMachine **********************************************/
/** Allocate resources needed by MSDOS-dependent code.      **/
/*************************************************************/
int InitMachine(void)
{
  union REGS R;
  struct SREGS SR;
  int I,J;
  static byte PalInit[16][3] =
  {
    {0x00,0x00,0x00},{0x00,0x00,0x00},{0x00,0xA0,0x00},{0x60,0xE0,0x60},
    {0x00,0x00,0x90},{0x00,0x00,0xF0},{0xA0,0x00,0x00},{0x00,0xC0,0xE0},
    {0xE0,0x40,0x40},{0xF0,0x80,0x80},{0xC0,0xA0,0x00},{0xE0,0xE0,0x60},
    {0x00,0x60,0x00},{0xE0,0x60,0xE0},{0x80,0x80,0x80},{0xE0,0xE0,0xE0},
  };

  if(Verbose)
    printf("Initializing IBM PC drivers:\n  Allocating screen buffer...");   

  XBuf=(byte *)malloc(sizeof(byte)*640*480);
  if(!XBuf) { if(Verbose) printf("FAILED\n");return(0); }
  if(Verbose) printf("OK\n  Allocating sprite buffer...");
  SBuf=(byte *)malloc(sizeof(byte)*256*212);
  if(!SBuf) { if(Verbose) printf("FAILED\n");return(0); }
  if(Verbose) printf("OK\n");

  /*** Set palette ***/
  outp(0x3C8,0);
  for(I=0;I<16;I++)
  {
    XPal[I]=I;
    for(J=0;J<3;J++)
      outp(0x3C9,PalInit[I][J]>>2);
  }
  XPal0=XPal[0];

  IRQ0Old=_dos_getvect(0x08);IRQ1Old=_dos_getvect(0x09);

  _dos_setvect(0x09,IRQ1Handler);

  if(!IntSync)
  {
    if(Verbose) printf("  Initializing timer for %dHz...",IFreq);

    Cmp=1193180L/IFreq/100;CntInt=100;RealClock=0L;

    cli();
    outp(0x43,0x36);outp(0x40,Cmp&0xFF);outp(0x40,(Cmp>>8)&0xFF);
    _dos_setvect(0x08,IRQ0Handler);
    sti();
  }

  return(1);
}


/** TrashMachine *********************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
  union REGS R;

  R.w.ax=0x0003;int386(0x10,&R,&R);                /* Set text mode 80x25 */

  if(Verbose) printf("Shutting down...\n");

  if(SBuf) free(SBuf);
  if(XBuf) free(XBuf);

  cli();
  _dos_setvect(0x08,IRQ0Old);                      /* Restore timer IRQ   */
  _dos_setvect(0x09,IRQ1Old);                      /* Restore keyboard    */
  outp(0x43,0x36);outp(0x40,0x00);outp(0x40,0x00); /* Set timer to 18.2Hz */
  sti();
}


/** InitScr256 ***********************************************/
/** Function to be called when 256-wide screen is set.      **/
/*************************************************************/
int InitScr256(void)
{
  union REGS R;

  /*** Set up 320x200x256 screen mode ***/
  R.w.ax=0x0013;int386(0x10,&R,&R);

  return(1);
}


/** ColorScrF ************************************************/
/** Generic function to be called when colors are changed.  **/
/*************************************************************/
void ColorsScrF(void)
{
  register byte C;
  register int J;

  outp(0x3C8,0);
  for(J=0;J<16;J++)
    for(C=0;C<3;C++)
      outp(0x3C9,Palette[J][C]>>2);

  C=XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];
  memset(XBuf,C,640*480);
}


/** InitTx80 *************************************************/
/** Function to be called when TEXT80 mode is set.          **/
/*************************************************************/
int InitTx80(void)
{ 
  union REGS R;
  
  /* Set up 80x25x16 text mode */
  R.w.ax=0x0003;int386(0x10,&R,&R);
  
  /* Turn off the cursor */
  outpw(0x03D4,0x1E0A);outpw(0x03D4,0x000B);

  return(1);
}   


/** ColorTx80 ************************************************/
/** Function to be called when colors are changed in TEXT80 **/
/*************************************************************/
void ColorsTx80(void)
{
  byte *P,C;
  int J;
    
  XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];

  outp(0x3C8,1);
  for(C=0;C<3;C++) outp(0x3C9,Palette[0][C]>>2);
  for(C=0;C<3;C++) outp(0x3C9,Palette[FGColor][C]>>2);
  for(C=0;C<3;C++) outp(0x3C9,Palette[BGColor][C]>>2);
  for(C=0;C<3;C++) outp(0x3C9,Palette[XFGColor][C]>>2);
  for(C=0;C<3;C++) outp(0x3C9,Palette[XBGColor][C]>>2);
        
  for(J=0,P=(byte *)0xB8000;J<80*25;J++) { *P++=' ';*P++=0x11; }  
}


/** InitScr512 ***********************************************/
/** Function to be called when 512-wide screen is set.      **/
/*************************************************************/
int InitScr512(void)
{
  union REGS R;
 
  /*** Set up 640x480x16 screen mode ***/
  R.w.ax=0x4F02;R.w.bx=0x0100;int386(0x10,&R,&R);
 
  return(1);
}   


/** Put640x480 ***********************************************/
/** Copy buffer contents to a 640x480 screen.               **/
/*************************************************************/
void Put640x480(byte *Buf)
{
  union REGS R;

  R.w.ax=0x4F05;R.w.bx=0;R.w.dx=0;int386(0x10,&R,&R);
  memcpy((void*)(0xA0000),(void*)(Buf+0x00000),0x10000);
  R.w.ax=0x4F05;R.w.bx=0;R.w.dx=1;int386(0x10,&R,&R);
  memcpy((void*)(0xA0000),(void*)(Buf+0x10000),0x10000);
  R.w.ax=0x4F05;R.w.bx=0;R.w.dx=2;int386(0x10,&R,&R);
  memcpy((void*)(0xA0000),(void*)(Buf+0x20000),0x10000);
  R.w.ax=0x4F05;R.w.bx=0;R.w.dx=3;int386(0x10,&R,&R);
  memcpy((void*)(0xA0000),(void*)(Buf+0x30000),0x10000);
}


/** Part of the code common for Unix/X and MSDOS drivers *****/
#include "Common.h"


/** RefreshScr6 **********************************************/
/** Function to be called to update SCREEN 6.               **/
/*************************************************************/
void RefreshScr6(register byte Y1,register byte Y2)
{
  register byte X,J,C;
  register byte *P,*T,*R;

  P=XBuf+(640-512)/2+Y1*640;

  if(ScreenON)
  {
    ColorSprites();

    R=SBuf+((word)Y1<<8);
    P+=WIDTH-256;
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<7);

    for(;Y1<=Y2;Y1++,J++,P+=2*(640-512)+512)
    {
      if(!J) T=ChrTab;
      for(X=0;X<128;X++,T++)
      {
        C=((*R>>4)|*R)&0x0F;
        *P=*(P+640)=XPal[C? C:(*T>>6)&0x03];P++;
        *P=*(P+640)=XPal[C? C:(*T>>4)&0x03];P++;
        R++;
        C=((*R>>4)|*R)&0x0F;
        *P=*(P+640)=XPal[C? C:(*T>>2)&0x03];P++;
        *P=*(P+640)=XPal[C? C:*T&0x03];P++;
        R++;
      }
    }
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH*2);

  if(EndOfFrame) Put640x480(XBuf);
}


/** RefreshScr7 **********************************************/
/** Function to be called to update SCREEN 7.               **/
/*************************************************************/
void RefreshScr7(register byte Y1,register byte Y2)
{
  register byte X,J,C;
  register byte *P,*T,*R;

  P=XBuf+(640-512)/2+Y1*640;

  if(ScreenON)
  {
    ColorSprites();

    R=SBuf+((word)Y1<<8);
    P+=WIDTH-256;
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<8);

    for(;Y1<=Y2;Y1++,J++,P+=2*(640-512)+512)
    {
      if(!J) T=ChrTab;
      for(X=0;X<128;X++)
      {
        C=((*R>>4)|*R)&0x0F;
        *P=*(P+640)=XPal[C? C:*T>>4];P++;
        *P=*(P+640)=XPal[C? C:*T&0x0F];P++;
        T++;R++;
        C=((*R>>4)|*R)&0x0F;
        *P=*(P+640)=XPal[C? C:*T>>4];P++;
        *P=*(P+640)=XPal[C? C:*T&0x0F];P++;
        T++;R++;
      }
    }
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH*2);

  if(EndOfFrame) Put640x480(XBuf);
}


/** RefreshScr8 **********************************************/
/** Function to be called to update SCREEN 8.               **/
/*************************************************************/
void RefreshScr8(register byte Y1,register byte Y2)
{
  register byte *P,*T,J;
  register word X;

  P=XBuf+WIDTH*(HEIGHT-192)/2+Y1*WIDTH;

  if(ScreenON)
  {
    P+=(WIDTH-256)/2;
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<8);

    for(;Y1<=Y2;Y1++,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<256;X++,T++) *P++=*T;
    }
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH);

  if(EndOfFrame) PutImage;
}


/** RefreshTx80 **********************************************/
/** Function to be called to update TEXT80.                 **/
/*************************************************************/
void RefreshTx80(register byte Y1,register byte Y2) 
{
  register byte X,Y,M;
  register byte *P,*C,*T;
  int Z;
    
  outpw(0x03C4,0x0402);outpw(0x03C4,0x0704);
  outpw(0x03CE,0x0204);outpw(0x03CE,0x0005);
  outpw(0x03CE,0x0406);

  P=(byte *)0xA0000;T=ChrGen;
  for(Z=0;Z<256;Z++,P+=16)
    for(Y=0;Y<8;Y++)
      *P++=*P++=(*T++)&0xFC;
  
  outpw(0x03C4,0x0302);outpw(0x03C4,0x0304);
  outpw(0x03CE,0x0004);outpw(0x03CE,0x1005);
  outpw(0x03CE,0x0A06);outpw(0x03CE,0x0E06);
      
  P=(byte *)0xB8000;
     
  if(ScreenON)
  {
    C=ColTab;T=ChrTab;M=0x80;
      
    for(Y=ScanLines212? 25:24;Y;Y--)
      for(X=0;X<80;X++)
      {
        *P++=*T++;
        *P++=(*C&M)? 0x32:0x54;
        if(!(M>>=1)) { M=0x80;C++; }
      }
  }
  else
    for(Z=80*(ScanLines212? 25:24);Z;Z--)
    { *P++=' ';*P++=0x33; }
}
