/** fMSX: portable MSX emulator ******************************/
/**                                                         **/
/**                           Unix.c                        **/
/**                                                         **/
/** This file contains Unix/X-dependent subroutines and     **/
/** drivers. It includes common drivers from Common.h.      **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1994,1995                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/


/** Standard Unix/X #includes ********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include "MSX.h"


/** MIT Shared Memory Extension for X ************************/
#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
XShmSegmentInfo SHMInfo;
int UseSHM=1;
#endif


/** Various X-related variables ******************************/
Screen *screen;
Display *disp;
Window root,win;
Colormap cmap;
XImage *ximage;
XEvent ev;
GC defaultGC,gc;
int screen_num;
unsigned long white,black;


/** Various variables and short functions ********************/
#define WIDTH  340
#define HEIGHT 240

char *Title="fMSX Unix/X 0.9";

int ExitFlag=0;
int IFreq=50;

struct itimerval TimerValue;

byte *XBuf,*SBuf;
byte BPal[256],XPal[16],XPal0,Pal[16][3];

void OnBreak(void) { ExitFlag=1; }
void OnTimer(void) { IFlag=1; }


/** PUTIMAGE: macro copying image buffer into a window ********/
#ifdef MITSHM
#define PutImage \
  if(UseSHM)                                                            \
    XShmPutImage(disp,win,defaultGC,ximage,0,0,0,0,WIDTH,HEIGHT,False); \
  else                                                                  \
    XPutImage(disp,win,defaultGC,ximage,0,0,0,0,WIDTH,HEIGHT);          \
  XFlush(disp)
#else
#define PutImage \
  XPutImage(disp,win,defaultGC,ximage,0,0,0,0,WIDTH,HEIGHT); \
  XFlush(disp);
#endif


/** Screen Mode Handlers [number of screens + 1] *************/
struct
{                                  /* Screen mode handlers:  */
  int (*Init)();                   /* Initialize screen      */
  void (*Colors)();                /* Update colors/palette  */
  void (*Refresh)(byte Y1,byte Y2);/* Refresh screen         */
  byte R2,R3,R4,R5;                /* Masks for table addr.  */
} SCR[MAXSCREEN+2] =               /* Number of screens + 1  */
{
  { InitScrF,ColorsScrF,RefreshScr0,0x7F,0x00,0x3F,0x00 }, /* SCR 0:TEXT 40x24  */
  { InitScrF,ColorsScrF,RefreshScr1,0x7F,0xFF,0x3F,0xFF }, /* SCR 1:TEXT 32x24  */
  { InitScrF,ColorsScrF,RefreshScr2,0x7F,0x80,0x3C,0xFF }, /* SCR 2:BLK 256x192 */
  { InitScrF,ColorsScrF,RefreshScr3,0x7F,0xFF,0x00,0xFF }, /* SCR 3:64x48x16    */
  { InitScrF,ColorsScrF,RefreshScr4,0x7F,0x80,0x3C,0xFC }, /* SCR 4:BLK 256x192 */
  { InitScrF,ColorsScrF,RefreshScr5,0x60,0x00,0x00,0xFC }, /* SCR 5:256x192x16  */
  { InitScrF,ColorsScrF,RefreshScr6,0x60,0x00,0x00,0xFC }, /* SCR 6:512x192x4   */
  { InitScrF,ColorsScrF,RefreshScr7,0x20,0x00,0x00,0xFC }, /* SCR 7:512x192x16  */
  { InitScrF,ColorsScrF,RefreshScr8,0x20,0x00,0x00,0xFC }, /* SCR 8:256x192x256 */
  { InitScrF,ColorsScrF,RefreshTx80,0x7C,0xF8,0x3F,0x00 }  /* SCR 0:TEXT 80x24  */
};


/** Keyboard bindings ****************************************/
byte Keys[256][2] =
{
/* 00 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {7,0x20},{7,0x08},{0,0x00},{0,0x00},{0,0x00},{7,0x80},{0,0x00},{0,0x00},
/* 10 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{7,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 20 */
  {8,0x01},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x01},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{2,0x04},{1,0x04},{2,0x08},{2,0x10},
/* 30 */
  {0,0x01},{0,0x02},{0,0x04},{0,0x08},{0,0x10},{0,0x20},{0,0x40},{0,0x80},
  {1,0x01},{1,0x02},{0,0x00},{1,0x80},{0,0x00},{1,0x08},{0,0x00},{0,0x00},
/* 40 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 50 */
  {0,0x00},{8,0x10},{8,0x20},{8,0x80},{8,0x40},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{1,0x20},{1,0x10},{1,0x40},{0,0x00},{0,0x00},
/* 60 */
  {2,0x02},{2,0x40},{2,0x80},{3,0x01},{3,0x02},{3,0x04},{3,0x08},{3,0x10},
  {3,0x20},{3,0x40},{3,0x80},{4,0x01},{4,0x02},{4,0x04},{4,0x08},{4,0x10},
/* 70 */
  {4,0x20},{4,0x40},{4,0x80},{5,0x01},{5,0x02},{5,0x04},{5,0x08},{5,0x10},
  {5,0x20},{5,0x40},{5,0x80},{0,0x00},{0,0x00},{0,0x00},{6,0x04},{0,0x00},
/* 80 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* 90 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* A0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* B0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{6,0x20},{6,0x40},
/* C0 */
  {6,0x80},{7,0x01},{7,0x02},{8,0x02},{8,0x08},{8,0x04},{7,0x40},{7,0x10},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* D0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {8,0x02},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* E0 */
  {0,0x00},{6,0x01},{6,0x08},{6,0x02},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{6,0x10},{6,0x04},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
/* F0 */
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},
  {0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{0,0x00},{8,0x08}
};


/** TrashMachine *********************************************/
/** Deallocate all resources taken by InitMachine().        **/
/*************************************************************/
void TrashMachine(void)
{
  if(Verbose) printf("Shutting down...\n");

  if(SBuf) free(SBuf);

  if(disp&&win)
  {
#ifdef MITSHM
    if(UseSHM)
    {
      XShmDetach(disp,&SHMInfo);
      if(SHMInfo.shmaddr) shmdt(SHMInfo.shmaddr);
      if(SHMInfo.shmid>=0) shmctl(SHMInfo.shmid,IPC_RMID,0);
    }
    else
#endif MITSHM
    if(ximage) XDestroyImage(ximage);
    XAutoRepeatOn(disp);
  }
  if(disp) XCloseDisplay(disp);
}


/** InitMachine **********************************************/
/** Allocate resources needed by Unix/X-dependent code.     **/
/*************************************************************/
int InitMachine(void)
{
  XGCValues values;
  int J;

  if(Verbose)
    printf("Initializing Unix/X drivers:\n  Opening display...");
  disp=XOpenDisplay(NULL);
  if(!disp) { if(Verbose) printf("FAILED\n");return(0); }

  screen=DefaultScreenOfDisplay(disp);screen_num=DefaultScreen(disp);
  white=XWhitePixel(disp,screen_num);black=XBlackPixel(disp,screen_num);
  root=DefaultRootWindow(disp);
  defaultGC=XDefaultGC(disp,screen_num);
  cmap=DefaultColormap(disp,screen_num);

  if(Verbose) printf("OK\n  Opening window...");
  win=XCreateSimpleWindow(disp,root,0,0,WIDTH,HEIGHT,0,white,black);
  if(!win) { if(Verbose) printf("FAILED\n");return(0); }

  {
    XSizeHints hints;
    XWMHints wmhints;

    hints.flags=USSize;hints.width=WIDTH;hints.height=HEIGHT;
    wmhints.input=True;wmhints.flags=InputHint;
    XSetWMHints(disp,win,&wmhints);
    XSetWMNormalHints(disp,win,&hints);
    XStoreName(disp,win,Title);
  }

  XSelectInput(disp,win,ExposureMask|KeyPressMask|KeyReleaseMask);
  XMapRaised(disp,win);XNextEvent(disp,&ev);

  gc=XCreateGC(disp,win,0,&values);
  XSetForeground(disp,gc,white);
  XSetBackground(disp,gc,black);
  XClearWindow(disp,win);
  XAutoRepeatOff(disp);

  if(Verbose) printf("OK\n  Allocating sprite buffer...");
  SBuf=(byte *)malloc(sizeof(byte)*256*212);
  if(!SBuf) { if(Verbose) printf("FAILED\n");return(0); }

#ifdef MITSHM
  if(UseSHM)
  {
    if(Verbose)
      printf("OK\n  Using shared memory:\n    Creating image...");
    ximage=XShmCreateImage
           (
             disp,DefaultVisual(disp,screen_num),8,
             ZPixmap,NULL,&SHMInfo,WIDTH,HEIGHT
           );
    if(!ximage) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n    Getting SHM info...");
    SHMInfo.shmid=shmget
                  (
                    IPC_PRIVATE,ximage->bytes_per_line*ximage->height,
                    IPC_CREAT|0777
                  );
    if(SHMInfo.shmid<0) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n    Allocating SHM...");
    XBuf=(byte *)(ximage->data=SHMInfo.shmaddr=shmat(SHMInfo.shmid,0,0));
    if(!XBuf) { if(Verbose) printf("FAILED\n");return(0); }

    SHMInfo.readOnly=False;
    if(Verbose) printf("OK\n    Attaching SHM...");
    if(!XShmAttach(disp,&SHMInfo))
    { if(Verbose) printf("FAILED\n");return(0); }
  }
  else
#endif
  {
    if(Verbose) printf("OK\n  Allocating screen buffer...");
    XBuf=(byte *)malloc(sizeof(byte)*HEIGHT*WIDTH);
    if(!XBuf) { if(Verbose) printf("FAILED\n");return(0); }

    if(Verbose) printf("OK\n  Creating image...");
    ximage=XCreateImage
           (
             disp,DefaultVisual(disp,screen_num),8,
             ZPixmap,0,XBuf,WIDTH,HEIGHT,8,0
           );
    if(!ximage) { if(Verbose) printf("FAILED\n");return(0); }
  }

  for(XPal0=black,J=0;J<16;J++)
  { memcpy(Pal[J],"\0\0\0",3);XPal[J]=black; }

  if(Verbose) printf("OK\n");

  if(!IntSync)
  {
    if(Verbose) printf("Initializing timer for %dHz...",IFreq);
    TimerValue.it_interval.tv_sec =
    TimerValue.it_value.tv_sec    = 0L;
    TimerValue.it_interval.tv_usec=
    TimerValue.it_value.tv_usec   = 1000000L/IFreq;
    J=setitimer(ITIMER_VIRTUAL,&TimerValue,NULL);
    if(Verbose) printf(J? "FAILED\n":"OK\n");
    if(J) return(0);
    signal(SIGVTALRM,OnTimer);
  }

  signal(SIGHUP,OnBreak);signal(SIGINT,OnBreak);
  signal(SIGQUIT,OnBreak);signal(SIGTERM,OnBreak);

  return(1);
}


/** Keyboard *************************************************/
/** Check for keyboard events, parse them, and modify MSX   **/
/** keyboard matrix.                                        **/
/*************************************************************/
void Keyboard(void)
{
  byte J;

  if(XCheckWindowEvent(disp,win,KeyPressMask|KeyReleaseMask,&ev))
  {
    J=XLookupKeysym((XKeyEvent *)&ev,0)&0xFF;
    if(ev.type==KeyPress)
    {
#ifdef DEBUG
      if(J==0xC9) Trace=!Trace;
#endif
      KeyMap[Keys[J][0]]&=~Keys[J][1];
    }
    else KeyMap[Keys[J][0]]|=Keys[J][1];
  }

  if(ExitFlag)
  { TrashMSX();TrashMachine();exit(0); }
}


/** InitScrF *************************************************/
/** Generic function to be called when screen is changed.   **/
/*************************************************************/
int InitScrF(void)
{
  static OldMode=0;
  XColor Color;
  register byte J,I;
  register word R,G,B;
  unsigned long L;
 
  if(Verbose&0x02) printf("Screen %d initialized\n",ScrMode);

  if(OldMode==8)
  { 
    for(R=0;R<32;R+=8)
      for(G=0;G<256;G+=64)
        for(B=0;B<4;B++)
        {
          L=BPal[R|G|B];
          if(L!=black) XFreeColors(disp,cmap,&L,1,0);
        }
  }

  if(ScrMode==8)
  {  
    for(R=0;R<4;R++)
      for(G=0;G<4;G++)
        for(B=0;B<4;B++)
        {
          Color.flags=DoRed|DoGreen|DoBlue;
          Color.red=(word)R<<14;
          Color.green=(word)G<<14;
          Color.blue=(word)B<<14;
          J=XAllocColor(disp,cmap,&Color)? Color.pixel:black;
          I=B|(R<<3)|(G<<6);
          BPal[I]=BPal[I|0x04]=BPal[I|0x20]=BPal[I|0x24]=J;
        }  
  }

  OldMode=ScrMode;return(1);
}


/** ColorScrF ************************************************/
/** Generic function to be called when colors are changed.  **/
/*************************************************************/
void ColorsScrF(void)
{
  XColor Color;
  unsigned long L;
  byte J;

  XPal[0]=XPal0;
  for(J=0;J<16;J++)
    if(memcmp(Palette[J],Pal[J],3))
    {
      memcpy(Pal[J],Palette[J],3);L=XPal[J];
      if(L!=black) XFreeColors(disp,cmap,&L,1,0);
      Color.flags=DoRed|DoGreen|DoBlue;
      Color.red=(word)Palette[J][0]<<8;
      Color.green=(word)Palette[J][1]<<8;
      Color.blue=(word)Palette[J][2]<<8;
      if(!memcmp(Pal[J],"\0\0\0",3)) XPal[J]=black;
      else
        if(XAllocColor(disp,cmap,&Color)) XPal[J]=Color.pixel;
        else { XPal[J]=black;memcpy(Pal[J],"\0\0\0",3); }
    }
  XPal0=XPal[0];

  XPal[0]=(!BGColor||SolidColor0)? XPal0:XPal[BGColor];

  L=ScanLine? WIDTH*(ScanLine+(HEIGHT-192)/2):0;
  memset(XBuf+L,XPal[0],WIDTH*HEIGHT-L);
}


/** Part of the code common for Unix/X and MSDOS drivers *****/ 
#include "Common.h"


/** RefreshScr6 **********************************************/
/** Function to be called to update SCREEN 6.               **/
/*************************************************************/
void RefreshScr6(register byte Y1,register byte Y2)
{
  register byte X,J,C;
  register byte *P,*T,*R;

  P=XBuf+WIDTH*(HEIGHT-192)/2+Y1*WIDTH;

  if(ScreenON)
  {
    ColorSprites();

    R=SBuf+((word)Y1<<8);
    P+=(WIDTH-256)/2;
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<7);

    for(;Y1<=Y2;Y1++,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<128;X++,T++)
      {
        C=((*R>>4)|*R)&0x0F;R++;*P++=XPal[C? C:*T>>6];
        C=((*R>>4)|*R)&0x0F;R++;*P++=XPal[C? C:(*T>>2)&0x03];
      }
    }  
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH);

  if(EndOfFrame) PutImage;
}
  

/** RefreshScr7 **********************************************/
/** Function to be called to update SCREEN 7.               **/
/*************************************************************/
void RefreshScr7(register byte Y1,register byte Y2)
{
  register byte X,J,C;
  register byte *P,*T,*R;

  P=XBuf+WIDTH*(HEIGHT-192)/2+Y1*WIDTH;

  if(ScreenON)
  {
    ColorSprites();

    R=SBuf+((word)Y1<<8);
    P+=(WIDTH-256)/2;
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<8);

    for(;Y1<=Y2;Y1++,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<128;X++)
      {
        C=((*R>>4)|*R)&0x0F;R++;*P++=XPal[C? C:*T>>4];T++;
        C=((*R>>4)|*R)&0x0F;R++;*P++=XPal[C? C:*T>>4];T++;
      }
    }
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH);

  if(EndOfFrame) PutImage;
}


/** RefreshScr7 **********************************************/
/** Function to be called to update SCREEN 7.               **/
/*************************************************************/
void RefreshScr8(register byte Y1,register byte Y2)
{
  register byte *P,*T,J;
  register word X;
  
  P=XBuf+WIDTH*(HEIGHT-192)/2+Y1*WIDTH;

  if(ScreenON)
  {
    P+=(WIDTH-256)/2;
    J=VDP[23]+Y1;
    T=ChrTab+((word)J<<8);

    for(;Y1<=Y2;Y1++,J++,P+=WIDTH-256)
    {
      if(!J) T=ChrTab;
      for(X=0;X<256;X++,T++) *P++=BPal[*T];
    }
  }
  else memset(P,XPal[BGColor],(Y2-Y1+1)*WIDTH);

  if(EndOfFrame) PutImage;
}


/** RefreshTx80 **********************************************/
/** Function to be called to update TEXT80.                 **/
/*************************************************************/
void RefreshTx80(register byte Y1,register byte Y2)
{
  register byte X,Y,J,K,M,FC,BC;
  register byte *P,*S,*C,*T;
  byte *G;

  P=XBuf+WIDTH*(HEIGHT-192)/2;
  G=FontBuf? FontBuf:ChrGen;

  if(ScreenON)
  {
    P+=(WIDTH-320)/2;C=ColTab;T=ChrTab;M=0x80;

    for(Y=ScanLines212? 26:24;Y;Y--,P+=WIDTH*8-4*80)
      for(X=0;X<80;X++,T++)
      {
        S=G+((long)*T<<3);
        if(*C&M) { FC=XPal[XFGColor];BC=XPal[XBGColor]; }
        else     { FC=XPal[FGColor];BC=XPal[BGColor]; }
        for(J=0;J<8;J++)
        {
          K=*S++;
          *P++=K&0xC0? FC:BC;
          *P++=K&0x30? FC:BC;
          *P++=K&0x0C? FC:BC;
          *P=BC;P+=WIDTH-3;
        }
        P+=4-WIDTH*8;
        if(!(M>>=1)) { M=0x80;C++; } 
      }
  }
  else memset(P+((word)Y1<<8),XPal[BGColor],(Y2-Y1+1)*WIDTH);
    
  if(EndOfFrame) PutImage;
}
